import json
from pathlib import Path

from models import PluginConfig
from poetry.core.pyproject.toml import PyProjectTOML
from tomlkit.toml_document import TOMLDocument

PLUGIN_NAME = "poetry-plugin-pycopy"
PROJECT_ROOT = Path(__name__).parent.absolute()
PROJECT_TOML_FILE = Path.joinpath(PROJECT_ROOT, "pyproject.toml")


class PoetryPluginPyCopyError(Exception):
    """PoetryPycopyPluginError"""


def create_line(k, v):
    """Create a line which will be written to a file.

    Example:
        __version = "1.0.4"
        __name = "My great app"

    """
    return f'__{k} = "{v}"\n'


def read_toml(toml_path: Path) -> TOMLDocument:
    """Read toml file

    Args:
        toml_path (Path): Path to .toml file

    Raises:
        PoetryPycopyPluginError: File not found

    Returns:
        TOMLDocument

    """

    # File not exists
    if not toml_path.exists():
        raise PoetryPluginPyCopyError(f"File not found {toml_path}")

    # Read file and return TOMLDocument object
    return PyProjectTOML(path=toml_path).data


def read_config(toml_data: TOMLDocument) -> PluginConfig:
    """Read [tool.poetry-pycopy-plugin] fields

    Args:
        toml_data: TOMLDocument

    Raises:
        PoetryPycopyPluginError: Field (key) not found in config section

    Returns:
        PluginConfig

    """

    # Verify that plugin config section exists in pyproject.toml
    try:
        toml_data["tool"][PLUGIN_NAME]
    except KeyError:
        raise PoetryPluginPyCopyError("Missing configuration in pyproject.toml")

    # Read config from pyproject.toml
    try:
        return PluginConfig(
            keys=list(toml_data["tool"][PLUGIN_NAME]["keys"]),
            dest_dir=toml_data["tool"][PLUGIN_NAME]["dest_dir"],
            dest_file=toml_data["tool"][PLUGIN_NAME]["dest_file"],
        )

    except KeyError as ex:
        raise PoetryPluginPyCopyError(ex)


def parse_fields(plugin_config: PluginConfig, toml_data: TOMLDocument) -> dict:
    """Parse requested keys from plugin config with keys from tool.poetry section.
    Only keys which exists in tool.poetry section will be returned.

    Args:
        plugin_config (PluginConfig): Plugin configuration
        toml_data (PyProjectTOML): Toml data

    Returns:
        dict: Parsed keys:values
    """

    # [tool.poetry] section
    tool_poetry = toml_data["tool"]["poetry"]

    # Find value in tool.poetry for every key from plugin configuration
    return {k: tool_poetry[k] for k in plugin_config.keys if k in tool_poetry}


def output_file_path(plugin_config: PluginConfig) -> Path:
    """Create path for output file from project root directory.

    Args:
        plugin_config (PluginConfig): Plugin configuration

    Returns:
        Path: Path to output file
    """
    return Path.joinpath(PROJECT_ROOT, plugin_config.dest_dir).joinpath(plugin_config.dest_file)


def pycopy():
    """Read pyproject.toml file from the project root, parse required fields and
    write them to the destination file.

    """

    print("\nRunning Poetry PyCopy Plugin:")

    toml_data: TOMLDocument = read_toml(toml_path=PROJECT_TOML_FILE)
    plugin_config: PluginConfig = read_config(toml_data=toml_data)

    # Destination file path
    dest_file_path = output_file_path(plugin_config)
    print("Destination file:", str(dest_file_path))

    # Parse
    parsed_data: dict = parse_fields(plugin_config, toml_data)
    print('\n', json.dumps(parsed_data, indent=2, default=str))

    # Create output-line for every record in parsed_data
    lines = [create_line(k, v) for k, v in parsed_data.items()]

    # Write lines to destination file
    with open(dest_file_path, "w+", encoding="utf-8") as f:
        f.writelines(lines)

    print("\n")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['poetry_plugin_pycopy']

package_data = \
{'': ['*']}

install_requires = \
['poetry>=1.2.2,<2.0.0']

entry_points = \
{'poetry.application.plugin': ['pycopy = plugin:PoetryPluginPycopy']}

setup_kwargs = {
    'name': 'poetry-plugin-pycopy',
    'version': '1.0.2',
    'description': 'Copy fields from pyproject.toml to source directory',
    'long_description': '# poetry-plugin-pycopy\n\n<!--- These are examples. See https://shields.io for others or to customize this set of shields. You might want to include dependencies, project status and licence info here --->\n![GitHub repo size](https://img.shields.io/github/repo-size/danielfajt/poetry-plugin-pycopy)\n![GitHub contributors](https://img.shields.io/github/contributors/danielfajt/poetry-plugin-pycopy)\n![GitHub stars](https://img.shields.io/github/stars/danielfajt/poetry-plugin-pycopy?style=social)\n![GitHub forks](https://img.shields.io/github/forks/danielfajt/poetry-plugin-pycopy?style=social)\n![Twitter Follow](https://img.shields.io/twitter/follow/danielfajt?style=social)\n\nThis plugin adds command `pycopy` to Poetry which will copy information from `pyproject.toml` to `source` directory. \n\nThe goal is to have `pyproject.toml` as a single source of truth for app version, name, description etc. and to have these values available during a program runtime.\n\n## Use case\nFastAPI app in which you want to show application name or version in API docs.\n\n\n## Installation\n\nFrom Pypi:\n```\n$ poetry self add poetry-plugin-pycopy\n```\n\n## Usage\n\n```\n$ poetry pycopy\n```\n\n## Plugin configuration in `pyproject.toml`\n\n```\n[tool.poetry-plugin-pycopy]\nkeys = ["name", "version", "description"]\ndest_dir = "<some_package_name>"\ndest_file = "__init__.py"\n```\n- `keys` list tells which fields should by copied from `[tool.poetry]`\n- `dest_dir` is package/module root\n- `dest_file` is the name of an output file\n\nPlugin also runs with `$poetry version` command automatically. So when you use version bump, e.g.: `$poetry version patch` the plugin will copy the new version value into the output file.\n\n## Output file example\nThe `dest_file` is set to `__init__.py`. Thus the plugin will create or replace that file with current values for a given `keys`. For example:\n\n```\n__name = "poetry-plugin-pycopy"\n__version = "1.0.0"\n__description = "Copy fields from pyproject.toml to source directory"\n\n```\n\n-> https://unlicense.org/',
    'author': 'Daniel Fajt',
    'author_email': 'daniel29se@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
